@@ -11,29 +11,32 @@
 use super::igd::{forward_port, IgdError};
 use super::wire_msg::WireMsg;
 use super::{
-    config::{Config, InternalConfig},
+    config::{Config, InternalConfig, RetryConfig, SERVER_NAME},
+    connection::{Connection, RecvStream, SendStream},
     connection_deduplicator::{ConnectionDeduplicator, DedupHandle},
-    connection_pool::{ConnId, ConnectionPool, ConnectionRemover},
-    connections::{
-        listen_for_incoming_connections, listen_for_incoming_messages, Connection,
-        DisconnectionEvents, RecvStream, SendStream,
+    connection_handle::{
+        listen_for_incoming_connections, listen_for_incoming_messages, ConnectionHandle,
+        DisconnectionEvents,
     },
+    connection_pool::{ConnId, ConnectionPool, ConnectionRemover},
     error::{
         ClientEndpointError, ConnectionError, EndpointError, RecvError, RpcError,
         SerializationError,
     },
 };
 use bytes::Bytes;
-use std::net::SocketAddr;
+use std::{
+    net::{IpAddr, SocketAddr},
+    sync::{
+        atomic::{AtomicUsize, Ordering},
+        Arc,
+    },
+};
 use tokio::sync::broadcast::{self, Sender};
 use tokio::sync::mpsc::{self, Receiver as MpscReceiver, Sender as MpscSender};
 use tokio::time::{timeout, Duration};
 use tracing::{debug, error, info, trace, warn};
 
-/// Host name of the Quic communication certificate used by peers
-// FIXME: make it configurable
-const CERT_SERVER_NAME: &str = "MaidSAFE.net";
-
 // Number of seconds before timing out the IGD request to forward a port.
 #[cfg(feature = "igd")]
 const PORT_FORWARD_TIMEOUT: Duration = Duration::from_secs(30);
@@ -46,24 +49,24 @@ const STANDARD_CHANNEL_SIZE: usize = 10000;
 
 /// Channel on which incoming messages can be listened to
 #[derive(Debug)]
-pub struct IncomingMessages<I: ConnId>(pub(crate) MpscReceiver<(Connection<I>, Bytes)>);
+pub struct IncomingMessages<I: ConnId>(pub(crate) MpscReceiver<(ConnectionHandle<I>, Bytes)>);
 
 impl<I: ConnId> IncomingMessages<I> {
     /// Blocks and returns the next incoming message and the connection it arrived on.
     ///
     /// **Note:** holding on to `Connection`
-    pub async fn next(&mut self) -> Option<(Connection<I>, Bytes)> {
+    pub async fn next(&mut self) -> Option<(ConnectionHandle<I>, Bytes)> {
         self.0.recv().await
     }
 }
 
 /// Channel on which incoming connections are notified on
-pub struct IncomingConnections<I: ConnId>(pub(crate) MpscReceiver<Connection<I>>);
+pub struct IncomingConnections<I: ConnId>(pub(crate) MpscReceiver<ConnectionHandle<I>>);
 
 impl<I: ConnId> IncomingConnections<I> {
     /// Blocks until there is an incoming connection and returns the address of the
     /// connecting peer
-    pub async fn next(&mut self) -> Option<Connection<I>> {
+    pub async fn next(&mut self) -> Option<ConnectionHandle<I>> {
         self.0.recv().await
     }
 }
@@ -74,12 +77,16 @@ pub struct Endpoint<I: ConnId> {
     local_addr: SocketAddr,
     public_addr: Option<SocketAddr>,
     quic_endpoint: quinn::Endpoint,
-    message_tx: MpscSender<(Connection<I>, Bytes)>,
+    retry_config: Arc<RetryConfig>,
+
+    message_tx: MpscSender<(ConnectionHandle<I>, Bytes)>,
     disconnection_tx: MpscSender<SocketAddr>,
-    config: InternalConfig,
     termination_tx: Sender<()>,
     connection_pool: ConnectionPool<I>,
     connection_deduplicator: ConnectionDeduplicator,
+
+    // counts fully opened connections, excluding incoming and connections dropped by connect_to_any
+    opened_connection_count: Arc<AtomicUsize>,
 }
 
 impl<I: ConnId> std::fmt::Debug for Endpoint<I> {
@@ -87,7 +94,7 @@ impl<I: ConnId> std::fmt::Debug for Endpoint<I> {
         f.debug_struct("Endpoint")
             .field("local_addr", &self.local_addr)
             .field("quic_endpoint", &"<endpoint omitted>")
-            .field("config", &self.config)
+            .field("retry_config", &self.retry_config)
             .finish()
     }
 }
@@ -127,7 +134,7 @@ impl<I: ConnId> Endpoint<I> {
             IncomingConnections<I>,
             IncomingMessages<I>,
             DisconnectionEvents,
-            Option<Connection<I>>,
+            Option<ConnectionHandle<I>>,
         ),
         EndpointError,
     > {
@@ -136,20 +143,26 @@ impl<I: ConnId> Endpoint<I> {
         let mut builder = quinn::Endpoint::builder();
         let _ = builder.listen(config.server.clone());
 
-        let (mut endpoint, quic_incoming, channels) =
-            Self::build_endpoint(local_addr.into(), config, builder)?;
+        let (mut endpoint, quic_incoming, channels) = Self::build_endpoint(
+            local_addr.into(),
+            config.client,
+            config.retry_config,
+            builder,
+        )?;
 
         let contact = endpoint.connect_to_any(contacts).await;
-        let public_addr = endpoint.resolve_public_addr(contact.as_ref()).await?;
+        let public_addr = endpoint
+            .resolve_public_addr(config.external_ip, config.external_port, contact.as_ref())
+            .await?;
 
         #[cfg(feature = "igd")]
-        if endpoint.config.forward_port {
+        if config.forward_port {
             timeout(
                 PORT_FORWARD_TIMEOUT,
                 forward_port(
                     public_addr.port(),
                     endpoint.local_addr(),
-                    endpoint.config.upnp_lease_duration,
+                    config.upnp_lease_duration,
                     channels.termination.1,
                 ),
             )
@@ -164,6 +177,7 @@ impl<I: ConnId> Endpoint<I> {
             channels.connection.0,
             channels.disconnection.0.clone(),
             endpoint.clone(),
+            endpoint.quic_endpoint.clone(),
         );
 
         if let Some(contact) = contact.as_ref() {
@@ -199,8 +213,12 @@ impl<I: ConnId> Endpoint<I> {
     ) -> Result<(Self, IncomingMessages<I>, DisconnectionEvents), ClientEndpointError> {
         let config = InternalConfig::try_from_config(config)?;
 
-        let (endpoint, _, channels) =
-            Self::build_endpoint(local_addr.into(), config, quinn::Endpoint::builder())?;
+        let (endpoint, _, channels) = Self::build_endpoint(
+            local_addr.into(),
+            config.client,
+            config.retry_config,
+            quinn::Endpoint::builder(),
+        )?;
 
         Ok((
             endpoint,
@@ -212,9 +230,11 @@ impl<I: ConnId> Endpoint<I> {
     // A private helper for initialising an endpoint.
     fn build_endpoint(
         local_addr: SocketAddr,
-        config: InternalConfig,
-        builder: quinn::EndpointBuilder,
+        client_config: quinn::ClientConfig,
+        retry_config: Arc<RetryConfig>,
+        mut builder: quinn::EndpointBuilder,
     ) -> Result<(Self, quinn::Incoming, Channels<I>), quinn::EndpointError> {
+        let _ = builder.default_client_config(client_config);
         let (quic_endpoint, quic_incoming) = builder.bind(&local_addr)?;
         let local_addr = quic_endpoint
             .local_addr()
@@ -226,12 +246,13 @@ impl<I: ConnId> Endpoint<I> {
             local_addr,
             public_addr: None,
             quic_endpoint,
+            retry_config,
             message_tx: channels.message.0.clone(),
             disconnection_tx: channels.disconnection.0.clone(),
-            config,
             termination_tx: channels.termination.0.clone(),
             connection_pool: ConnectionPool::new(),
             connection_deduplicator: ConnectionDeduplicator::new(),
+            opened_connection_count: Arc::new(AtomicUsize::new(0)),
         };
 
         Ok((endpoint, quic_incoming, channels))
@@ -247,6 +268,16 @@ impl<I: ConnId> Endpoint<I> {
         self.public_addr.unwrap_or(self.local_addr)
     }
 
+    /// Get the count of opened connections.
+    ///
+    /// `Endpoint`s keep a count of connections they have fully opened and returned to callers.
+    /// Notably this excludes any incoming connections, and connections that were dropped by
+    /// [`connect_to_any`](Self::connect_to_any).
+    #[stability::unstable(feature = "opened-connection-count")]
+    pub fn opened_connection_count(&self) -> usize {
+        self.opened_connection_count.load(Ordering::Relaxed)
+    }
+
     /// Removes all existing connections to a given peer
     pub async fn disconnect_from(&self, peer_addr: &SocketAddr) {
         self.connection_pool
@@ -254,7 +285,7 @@ impl<I: ConnId> Endpoint<I> {
             .await
             .iter()
             .for_each(|conn| {
-                conn.close(0u8.into(), b"");
+                conn.close();
             });
     }
 
@@ -282,7 +313,7 @@ impl<I: ConnId> Endpoint<I> {
     pub async fn connect_to(
         &self,
         node_addr: &SocketAddr,
-    ) -> Result<Connection<I>, ConnectionError> {
+    ) -> Result<ConnectionHandle<I>, ConnectionError> {
         self.get_or_connect_to(node_addr).await
     }
 
@@ -301,7 +332,7 @@ impl<I: ConnId> Endpoint<I> {
     /// method will check the pool before opening each connection. If a new connection is opened, it
     /// will be added to the pool. Note that already pooled connections will have a higher chance of
     /// 'winning' the race, and being the selected peer.
-    pub async fn connect_to_any(&self, peer_addrs: &[SocketAddr]) -> Option<Connection<I>> {
+    pub async fn connect_to_any(&self, peer_addrs: &[SocketAddr]) -> Option<ConnectionHandle<I>> {
         trace!("Connecting to any of {:?}", peer_addrs);
         if peer_addrs.is_empty() {
             return None;
@@ -335,12 +366,13 @@ impl<I: ConnId> Endpoint<I> {
         trace!("Checking is reachable");
 
         // avoid the connection pool
-        let quinn::NewConnection { connection, .. } = self.new_connection(peer_addr).await?;
-        let (send_stream, recv_stream) = connection.open_bi().await?;
-        let mut send_stream = SendStream::new(send_stream);
-        let mut recv_stream = RecvStream::new(recv_stream);
+        let (connection, _) = Connection::new(
+            self.quic_endpoint.clone(),
+            self.new_connection(peer_addr).await?,
+        );
+        let (mut send_stream, mut recv_stream) = connection.open_bi().await?;
 
-        send_stream.send(WireMsg::EndpointEchoReq).await?;
+        send_stream.send_wire_msg(WireMsg::EndpointEchoReq).await?;
 
         match timeout(ECHO_SERVICE_QUERY_TIMEOUT, recv_stream.next_wire_msg()).await?? {
             Some(WireMsg::EndpointEchoResp(_)) => Ok(()),
@@ -362,7 +394,7 @@ impl<I: ConnId> Endpoint<I> {
     }
 
     /// Get the existing `Connection` for a `SocketAddr`.
-    pub async fn get_connection_by_addr(&self, addr: &SocketAddr) -> Option<Connection<I>> {
+    pub async fn get_connection_by_addr(&self, addr: &SocketAddr) -> Option<ConnectionHandle<I>> {
         self.connection_pool
             .get_by_addr(addr)
             .await
@@ -370,7 +402,7 @@ impl<I: ConnId> Endpoint<I> {
     }
 
     /// Get the existing `Connection` for the given ID.
-    pub async fn get_connection_by_id(&self, id: &I) -> Option<Connection<I>> {
+    pub async fn get_connection_by_id(&self, id: &I) -> Option<ConnectionHandle<I>> {
         self.connection_pool
             .get_by_id(id)
             .await
@@ -413,7 +445,7 @@ impl<I: ConnId> Endpoint<I> {
     pub(crate) async fn get_or_connect_to(
         &self,
         addr: &SocketAddr,
-    ) -> Result<Connection<I>, ConnectionError> {
+    ) -> Result<ConnectionHandle<I>, ConnectionError> {
         let completion = loop {
             if let Some((conn, remover)) = self.connection_pool.get_by_addr(addr).await {
                 trace!("We are already connected to this peer: {}", addr);
@@ -432,7 +464,8 @@ impl<I: ConnId> Endpoint<I> {
             Ok(new_connection) => {
                 trace!("Successfully connected to peer: {}", addr);
 
-                let connection = new_connection.connection;
+                let (connection, connection_incoming) =
+                    Connection::new(self.quic_endpoint.clone(), new_connection);
                 let id = ConnId::generate(&connection.remote_address());
                 let remover = self
                     .connection_pool
@@ -442,14 +475,15 @@ impl<I: ConnId> Endpoint<I> {
 
                 listen_for_incoming_messages(
                     connection.clone(),
-                    new_connection.uni_streams,
-                    new_connection.bi_streams,
+                    connection_incoming,
                     self.message_tx.clone(),
                     self.disconnection_tx.clone(),
-                    self.clone(),
                 );
 
                 let _ = completion.complete(Ok(()));
+
+                let _ = self.opened_connection_count.fetch_add(1, Ordering::Relaxed);
+
                 Ok(connection)
             }
             Err(error) => {
@@ -467,15 +501,10 @@ impl<I: ConnId> Endpoint<I> {
         &self,
         node_addr: &SocketAddr,
     ) -> Result<quinn::NewConnection, ConnectionError> {
-        self.config
-            .retry_config
+        self.retry_config
             .retry(|| async {
                 trace!("Attempting to connect to {:?}", node_addr);
-                let connecting = match self.quic_endpoint.connect_with(
-                    self.config.client.clone(),
-                    node_addr,
-                    CERT_SERVER_NAME,
-                ) {
+                let connecting = match self.quic_endpoint.connect(node_addr, SERVER_NAME) {
                     Ok(conn) => Ok(conn),
                     Err(error) => {
                         warn!("Connection attempt failed due to {:?}", error);
@@ -502,7 +531,9 @@ impl<I: ConnId> Endpoint<I> {
     // set an appropriate public address based on `config` and a reachability check.
     async fn resolve_public_addr(
         &mut self,
-        contact: Option<&Connection<I>>,
+        config_external_ip: Option<IpAddr>,
+        config_external_port: Option<u16>,
+        contact: Option<&ConnectionHandle<I>>,
     ) -> Result<SocketAddr, EndpointError> {
         let mut public_addr = self.local_addr;
 
@@ -518,7 +549,7 @@ impl<I: ConnId> Endpoint<I> {
             None
         };
 
-        if let Some(external_ip) = self.config.external_ip {
+        if let Some(external_ip) = config_external_ip {
             // set the public IP based on config
             public_addr.set_ip(external_ip);
 
@@ -543,7 +574,7 @@ impl<I: ConnId> Endpoint<I> {
             );
         }
 
-        if let Some(external_port) = self.config.external_port {
+        if let Some(external_port) = config_external_port {
             // set the public port based on config
             public_addr.set_port(external_port);
 
@@ -575,10 +606,10 @@ impl<I: ConnId> Endpoint<I> {
     }
 
     /// Perform the endpoint echo RPC with the given contact.
-    async fn endpoint_echo(&self, contact: &Connection<I>) -> Result<SocketAddr, RpcError> {
+    async fn endpoint_echo(&self, contact: &ConnectionHandle<I>) -> Result<SocketAddr, RpcError> {
         let (mut send, mut recv) = contact.open_bi(0).await?;
 
-        send.send(WireMsg::EndpointEchoReq).await?;
+        send.send_wire_msg(WireMsg::EndpointEchoReq).await?;
 
         match timeout(ECHO_SERVICE_QUERY_TIMEOUT, recv.next_wire_msg()).await?? {
             Some(WireMsg::EndpointEchoResp(addr)) => Ok(addr),
@@ -589,12 +620,12 @@ impl<I: ConnId> Endpoint<I> {
     /// Perform the endpoint verification RPC with the given peer.
     async fn endpoint_verification(
         &self,
-        contact: &Connection<I>,
+        contact: &ConnectionHandle<I>,
         public_addr: SocketAddr,
     ) -> Result<bool, RpcError> {
         let (mut send, mut recv) = contact.open_bi(0).await?;
 
-        send.send(WireMsg::EndpointVerificationReq(public_addr))
+        send.send_wire_msg(WireMsg::EndpointVerificationReq(public_addr))
             .await?;
 
         match timeout(ECHO_SERVICE_QUERY_TIMEOUT, recv.next_wire_msg()).await?? {
@@ -606,17 +637,20 @@ impl<I: ConnId> Endpoint<I> {
     /// Wrap a quinn connection, setting the default retry config and pool remover.
     pub(crate) fn wrap_connection(
         &self,
-        connection: quinn::Connection,
+        connection: Connection,
         remover: ConnectionRemover<I>,
-    ) -> Connection<I> {
-        Connection::new(connection, self.config.retry_config.clone(), remover)
+    ) -> ConnectionHandle<I> {
+        ConnectionHandle::new(connection, self.retry_config.clone(), remover)
     }
 }
 
 // a private helper struct for passing a bunch of channel-related things
-type Msg<I> = (Connection<I>, Bytes);
+type Msg<I> = (ConnectionHandle<I>, Bytes);
 struct Channels<I: ConnId> {
-    connection: (MpscSender<Connection<I>>, MpscReceiver<Connection<I>>),
+    connection: (
+        MpscSender<ConnectionHandle<I>>,
+        MpscReceiver<ConnectionHandle<I>>,
+    ),
     message: (MpscSender<Msg<I>>, MpscReceiver<Msg<I>>),
     disconnection: (MpscSender<SocketAddr>, MpscReceiver<SocketAddr>),
     termination: (Sender<()>, broadcast::Receiver<()>),
