@@ -28,9 +28,12 @@ async fn successful_connection() -> Result<()> {
     if let Some(connection) = peer1_incoming_connections.next().await {
         assert_eq!(connection.remote_address(), peer2_addr);
     } else {
-        eyre!("No incoming connection");
+        bail!("No incoming connection");
     }
 
+    assert_eq!(peer1.opened_connection_count(), 0);
+    assert_eq!(peer2.opened_connection_count(), 1);
+
     Ok(())
 }
 
@@ -52,7 +55,7 @@ async fn single_message() -> Result<()> {
     if let Some(connection) = peer1_incoming_connections.next().await {
         assert_eq!(connection.remote_address(), peer2_addr);
     } else {
-        eyre!("No incoming connection");
+        bail!("No incoming connection");
     }
 
     // Peer 2 gets an incoming message
@@ -60,8 +63,12 @@ async fn single_message() -> Result<()> {
         assert_eq!(connection.remote_address(), peer2_addr);
         assert_eq!(message, msg_from_peer2);
     } else {
-        eyre!("No incoming message");
+        bail!("No incoming message");
     }
+
+    assert_eq!(peer1.opened_connection_count(), 0);
+    assert_eq!(peer2.opened_connection_count(), 1);
+
     Ok(())
 }
 
@@ -83,14 +90,14 @@ async fn reuse_outgoing_connection() -> Result<()> {
     if let Some(connection) = bob_incoming_connections.next().await {
         assert_eq!(connection.remote_address(), alice_addr);
     } else {
-        eyre!("No incoming connection");
+        bail!("No incoming connection");
     }
 
     if let Some((connection, message)) = bob_incoming_messages.next().await {
         assert_eq!(connection.remote_address(), alice_addr);
         assert_eq!(message, msg0);
     } else {
-        eyre!("No incoming message");
+        bail!("No incoming message");
     }
 
     // Try connecting again and send a message
@@ -102,7 +109,7 @@ async fn reuse_outgoing_connection() -> Result<()> {
     if let Ok(Some(connection)) =
         timeout(Duration::from_secs(2), bob_incoming_connections.next()).await
     {
-        eyre!(
+        bail!(
             "Unexpected incoming connection from {}",
             connection.remote_address()
         );
@@ -112,8 +119,12 @@ async fn reuse_outgoing_connection() -> Result<()> {
         assert_eq!(connection.remote_address(), alice_addr);
         assert_eq!(message, msg1);
     } else {
-        eyre!("No incoming message");
+        bail!("No incoming message");
     }
+
+    assert_eq!(alice.opened_connection_count(), 1);
+    assert_eq!(bob.opened_connection_count(), 0);
+
     Ok(())
 }
 
@@ -136,14 +147,14 @@ async fn reuse_incoming_connection() -> Result<()> {
     if let Some(connection) = bob_incoming_connections.next().await {
         assert_eq!(connection.remote_address(), alice_addr);
     } else {
-        eyre!("No incoming connection");
+        bail!("No incoming connection");
     }
 
     if let Some((connection, message)) = bob_incoming_messages.next().await {
         assert_eq!(connection.remote_address(), alice_addr);
         assert_eq!(message, msg0);
     } else {
-        eyre!("No incoming message");
+        bail!("No incoming message");
     }
 
     // Bob tries to connect to alice and sends a message
@@ -156,7 +167,7 @@ async fn reuse_incoming_connection() -> Result<()> {
     if let Ok(Some(connection)) =
         timeout(Duration::from_secs(2), alice_incoming_connections.next()).await
     {
-        eyre!(
+        bail!(
             "Unexpected incoming connection from {}",
             connection.remote_address()
         );
@@ -166,8 +177,12 @@ async fn reuse_incoming_connection() -> Result<()> {
         assert_eq!(connection.remote_address(), bob_addr);
         assert_eq!(message, msg1);
     } else {
-        eyre!("No incoming message");
+        bail!("No incoming message");
     }
+
+    assert_eq!(alice.opened_connection_count(), 1);
+    assert_eq!(bob.opened_connection_count(), 0);
+
     Ok(())
 }
 
@@ -186,7 +201,7 @@ async fn disconnection() -> Result<()> {
     if let Some(connection) = bob_incoming_connections.next().await {
         assert_eq!(connection.remote_address(), alice_addr);
     } else {
-        eyre!("No incoming connection");
+        bail!("No incoming connection");
     }
 
     // After Alice disconnects from Bob both peers should receive the disconnected event.
@@ -195,13 +210,13 @@ async fn disconnection() -> Result<()> {
     if let Some(disconnected_peer) = alice_disconnections.next().await {
         assert_eq!(disconnected_peer, bob_addr);
     } else {
-        eyre!("Missing disconnection event");
+        bail!("Missing disconnection event");
     }
 
     if let Some(disconnected_peer) = bob_disconnections.next().await {
         assert_eq!(disconnected_peer, alice_addr);
     } else {
-        eyre!("Missing disconnection event");
+        bail!("Missing disconnection event");
     }
 
     // This time bob connects to Alice. Since this is a *new connection*, Alice should get the connection event
@@ -210,9 +225,12 @@ async fn disconnection() -> Result<()> {
     if let Some(connection) = alice_incoming_connections.next().await {
         assert_eq!(connection.remote_address(), bob_addr);
     } else {
-        eyre!("Missing incoming connection");
+        bail!("Missing incoming connection");
     }
 
+    assert_eq!(alice.opened_connection_count(), 1);
+    assert_eq!(bob.opened_connection_count(), 1);
+
     Ok(())
 }
 
@@ -241,13 +259,13 @@ async fn simultaneous_incoming_and_outgoing_connections() -> Result<()> {
     if let Some(connection) = alice_incoming_connections.next().await {
         assert_eq!(connection.remote_address(), bob_addr);
     } else {
-        eyre!("No incoming connection from Bob");
+        bail!("No incoming connection from Bob");
     }
 
     if let Some(connection) = bob_incoming_connections.next().await {
         assert_eq!(connection.remote_address(), alice_addr);
     } else {
-        eyre!("No incoming connectino from Alice");
+        bail!("No incoming connectino from Alice");
     }
 
     let msg0 = random_msg(1024);
@@ -260,37 +278,35 @@ async fn simultaneous_incoming_and_outgoing_connections() -> Result<()> {
         assert_eq!(connection.remote_address(), bob_addr);
         assert_eq!(message, msg1);
     } else {
-        eyre!("Missing incoming message");
+        bail!("Missing incoming message");
     }
 
     if let Some((connection, message)) = bob_incoming_messages.next().await {
         assert_eq!(connection.remote_address(), alice_addr);
         assert_eq!(message, msg0);
     } else {
-        eyre!("Missing incoming message");
+        bail!("Missing incoming message");
     }
 
-    // Drop the connection initiated by Bob.
+    // Drop all connections between Bob and Alice.
     bob.disconnect_from(&alice_addr).await;
 
     // It should be closed on Alice's side too.
     if let Some(disconnected_peer) = alice_disconnections.next().await {
         assert_eq!(disconnected_peer, bob_addr);
     } else {
-        eyre!("Missing disconnection event");
+        bail!("Missing disconnection event");
     }
 
-    // Bob connects to Alice again. This does not open a new connection but returns the connection
-    // previously initiated by Alice from the pool.
+    // Bob connects to Alice again. This opens a new connection.
     let b_to_a = bob.connect_to(&alice_addr).await?;
 
     if let Ok(Some(connection)) =
-        timeout(Duration::from_secs(2), alice_incoming_connections.next()).await
+        timeout(Duration::from_secs(60), alice_incoming_connections.next()).await
     {
-        eyre!(
-            "Unexpected incoming connection from {}",
-            connection.remote_address()
-        );
+        assert_eq!(connection.remote_address(), bob_addr);
+    } else {
+        bail!("Missing incoming connection");
     }
 
     let msg2 = random_msg(1024);
@@ -301,6 +317,9 @@ async fn simultaneous_incoming_and_outgoing_connections() -> Result<()> {
         assert_eq!(message, msg2);
     }
 
+    assert_eq!(alice.opened_connection_count(), 1);
+    assert_eq!(bob.opened_connection_count(), 2);
+
     Ok(())
 }
 
@@ -322,13 +341,13 @@ async fn multiple_concurrent_connects_to_the_same_peer() -> Result<()> {
     if let Some(connection) = alice_incoming_connections.next().await {
         assert_eq!(connection.remote_address(), bob_addr);
     } else {
-        eyre!("Missing incoming connection");
+        bail!("Missing incoming connection");
     }
 
     if let Ok(Some(connection)) =
         timeout(Duration::from_secs(2), alice_incoming_connections.next()).await
     {
-        eyre!(
+        bail!(
             "Unexpected incoming connection from {}",
             connection.remote_address()
         );
@@ -347,7 +366,7 @@ async fn multiple_concurrent_connects_to_the_same_peer() -> Result<()> {
 
     // Bob did not get a new incoming connection
     if let Ok(Some(_)) = timeout(Duration::from_secs(2), bob_incoming_connections.next()).await {
-        eyre!("Unexpected incoming connection from alice to bob");
+        bail!("Unexpected incoming connection from alice to bob");
     }
 
     // Both messages are received  at the other end
@@ -355,16 +374,19 @@ async fn multiple_concurrent_connects_to_the_same_peer() -> Result<()> {
         assert_eq!(connection.remote_address(), bob_addr);
         assert_eq!(message, msg1);
     } else {
-        eyre!("No message received from Bob");
+        bail!("No message received from Bob");
     }
 
     if let Some((connection, message)) = bob_incoming_messages.next().await {
         assert_eq!(connection.remote_address(), alice_addr);
         assert_eq!(message, msg0);
     } else {
-        eyre!("No message from alice");
+        bail!("No message from alice");
     }
 
+    assert_eq!(alice.opened_connection_count(), 0);
+    assert_eq!(bob.opened_connection_count(), 1);
+
     Ok(())
 }
 
@@ -461,12 +483,17 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
                     }
                 }
 
+                assert_eq!(send_endpoint.opened_connection_count(), 1);
+
                 Ok::<_, Report>(())
             }
         }));
     }
 
     let _ = future::try_join_all(tasks).await?;
+
+    assert_eq!(server_endpoint.opened_connection_count(), 0);
+
     Ok(())
 }
 
@@ -569,6 +596,8 @@ async fn multiple_connections_with_many_larger_concurrent_messages() -> Result<(
                     }
                 }
 
+                assert_eq!(send_endpoint.opened_connection_count(), 1);
+
                 Ok::<_, Report>(())
             }
         }));
@@ -580,6 +609,9 @@ async fn multiple_connections_with_many_larger_concurrent_messages() -> Result<(
             other => bail!("Error from test threads: {:?}", other??),
         }
     }
+
+    assert_eq!(server_endpoint.opened_connection_count(), 0);
+
     Ok(())
 }
 
@@ -637,6 +669,10 @@ async fn many_messages() -> Result<()> {
     }));
 
     let _ = future::try_join_all(tasks).await?;
+
+    assert_eq!(send_endpoint.opened_connection_count(), 1);
+    assert_eq!(recv_endpoint.opened_connection_count(), 0);
+
     Ok(())
 }
 
@@ -671,6 +707,12 @@ async fn connection_attempts_to_bootstrap_contacts_should_succeed() -> Result<()
             ep.connect_to(&peer).await.map(drop)?;
         }
     }
+
+    assert_eq!(ep1.opened_connection_count(), 0);
+    assert_eq!(ep2.opened_connection_count(), 0);
+    assert_eq!(ep3.opened_connection_count(), 0);
+    assert_eq!(ep.opened_connection_count(), 3);
+
     Ok(())
 }
 
@@ -680,10 +722,14 @@ async fn reachability() -> Result<()> {
     let (ep2, _, _, _, _) = new_endpoint().await?;
 
     if let Ok(()) = ep1.is_reachable(&"127.0.0.1:12345".parse()?).await {
-        eyre!("Unexpected success");
+        bail!("Unexpected success");
     };
     let reachable_addr = ep2.public_addr();
     ep1.is_reachable(&reachable_addr).await?;
+
+    assert_eq!(ep1.opened_connection_count(), 0);
+    assert_eq!(ep2.opened_connection_count(), 0);
+
     Ok(())
 }
 
@@ -733,5 +779,8 @@ async fn client() -> Result<()> {
         .ok_or_else(|| eyre!("Did not receive expected disconnection"))?;
     assert_eq!(disconnector, server.public_addr());
 
+    assert_eq!(server.opened_connection_count(), 0);
+    assert_eq!(client.opened_connection_count(), 1);
+
     Ok(())
 }
